#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
from bdgenomics.mango.coverage import *
from bdgenomics.mango.test import SparkTestCase
from collections import Counter

from bdgenomics.adam.adamContext import ADAMContext


class CoverageTest(SparkTestCase):

    def test_coverage_distribution(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments

        reads = ac.loadAlignments(testFile)

        # convert to coverage
        coverage = reads.toCoverage()

        qc = CoverageDistribution(self.ss, coverage, bin_size = 1)

        _, cd = qc.plotDistributions(testMode = True, normalize=False)

        assert(len(cd) == 1)

        # all items for first sample
        items = list(cd.popitem()[1])

        assert(items[0][1] == 1500)

    def test_example_coverage(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.exampleFile("chr17.7500000-7515000.sam")
        # read alignments
        alignments = ac.loadAlignments(testFile)
        coverage = alignments.toCoverage()

        qc = CoverageDistribution(self.ss, coverage, bin_size = 1)
        # sum of all coverage

        _, cd1 = qc.plotDistributions(testMode = True, cumulative = False, normalize=False)
        total = sum(map(lambda x: x[1], list(qc.collectedCounts.items())[0][1]))

        # first sample
        items = list(cd1.popitem()[1])
        x = items[0]
        assert(x[0] == 1) # 6 locations with read depth 1
        assert(x[1] == 6)

        _, cd2 = qc.plotDistributions(testMode = True, cumulative = False, normalize=True)

        # first sample
        items = list(cd2.popitem()[1])
        x = items[0]
        assert(x[0] == 1)
        assert(x[1] == 6.0/total) # normalized value

        _, cd3 = qc.plotDistributions(testMode = True, cumulative = True, normalize = True)

        # first sample
        items = list(cd3.popitem()[1])
        x = items[-1]
        assert(x[0] == 89)
        assert(x[1] > 0.999) # cumulative and normalized, so last value shound be about 1



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from bdgenomics.mango.test import SparkTestCase
from bdgenomics.mango.features import *

from bdgenomics.adam.adamContext import ADAMContext


class FeatureTest(SparkTestCase):

    def test_visualize_features(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("smalltest.bed")

        # read features
        features = ac.loadFeatures(testFile)

        featureViz = FeatureSummary(ac, features)

        contig = "chrM"
        start = 1
        end = 2000

        x = featureViz.viewPileup(contig, start, end)
        assert(x != None)



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

r"""
=========
Genotypes
=========
.. currentmodule:: bdgenomics.mango.genotypes
.. autosummary::
   :toctree: _generate/

   GenotypeSummary
   VariantsPerSampleDistribution
"""

from bdgenomics.mango.pileup.track import *

from .distribution import HistogramDistribution
import matplotlib.pyplot as plt
plt.rcdefaults()

class GenotypeSummary(object):
    """GenotypeSummary class.
    GenotypeSummary provides visualizations for genotype based summaries.
    """

    def __init__(self, spark, ac, genotypeRdd, sample = 1.0):
        """
        Initializes an GenotypeSummary class.

        Args:
            :param spark: SparkSession
            :param ac: ADAMContext
            :param genotypeRdd: GenotypeRDD
            :param sample: fraction of reads to sample from
        """

        # sample must be between 0 and 1
        if sample <= 0 or sample > 1:
            raise Exception('sample {} should be > 0 and <= 1'.format(self.sample))

        self.ss = spark
        self.ac = ac
        self.sample = sample
        # sample rdd
        self.genotypeRdd = genotypeRdd

        # where to store collected data
        self.variantsPerSampleDistribution = None

    def getVariantsPerSampleDistribution(self):
        """
        Computes distribution of variant counts per sample.

        Returns:
           VariantsPerSampleDistribution object
        """
        if self.variantsPerSampleDistribution == None:
            print("Computing coverage variants per sample distribution...")
            # pre-sample before computing coverage

            self.variantsPerSampleDistribution = VariantsPerSampleDistribution(self.ss, \
                                                                               self.genotypeRdd, \
                                                                               sample = self.sample)

        return self.variantsPerSampleDistribution


class VariantsPerSampleDistribution(HistogramDistribution):
    """ VariantsPerSampleDistribution class.
    VariantsPerSampleDistribution computes a distribution of the count of variants per sample.
    """

    def __init__(self, ss, genotypeRDD, sample = 1.0):
        """
        Initializes a VariantsPerSampleDistributionn class.
        Computes the coverage distribution of a CoverageRDD. This RDD can have data for multiple samples.

        Args:
            :param ss: global SparkSession.
            :param genotypeRDD: A bdgenomics.adam.rdd.GenotypeRDD object.
            :param sample: Fraction to sample CoverageRDD. Should be between 0 and 1
        """

        self.sc = ss.sparkContext
        self.sample = sample

        self.rdd = genotypeRDD.toDF().rdd \
            .filter(lambda r: ( (r.alleles[0] == u'ALT') | (r.alleles[1] == u'ALT'))) \
            .map(lambda r: ((r["sampleId"]), 1))

        HistogramDistribution.__init__(self)



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from bdgenomics.mango.coverage import CoverageDistribution
from bdgenomics.mango.test import SparkTestCase

from bdgenomics.adam.adamContext import ADAMContext


class DistributionTest(SparkTestCase):

    def test_normalized_count_distribution(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments

        reads = ac.loadAlignments(testFile)

        # convert to coverage
        coverage = reads.toCoverage()

        qc = CoverageDistribution(self.ss, coverage)

        _, cd = qc.plotDistributions(testMode = True, normalize = True)
        items = list(cd.popitem()[1])
        assert(len(items) == 1)
        assert(items.pop()[1] == 1.0)

        _, cd = qc.plotDistributions(testMode = True, normalize = False)
        items = list(cd.popitem()[1])
        assert(len(items) == 1)
        assert(items.pop()[1] == 1500)


    def test_cumulative_count_distribution(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments

        reads = ac.loadAlignments(testFile)

        # convert to coverage
        coverage = reads.toCoverage()

        qc = CoverageDistribution(self.ss, coverage)

        _, cd = qc.plotDistributions(testMode = True, cumulative = True, normalize = False)

        # first sample
        items = list(cd.popitem()[1])
        assert(len(items) == 1)
        assert(items.pop()[1] == 1500)

        _, cd = qc.plotDistributions(testMode = True, cumulative = False, normalize = False)

        # first sample
        items = list(cd.popitem()[1])
        assert(len(items) == 1)
        assert(items.pop()[1] == 1500)

    def test_fail_on_invalid_sample(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments

        reads = ac.loadAlignments(testFile)

        # convert to coverage
        coverage = reads.toCoverage()

        with self.assertRaises(Exception):
            CoverageDistribution(self.ss, coverage, sample = 1.2)
            CoverageDistribution(self.ss, coverage, sample = 0)

    def test_sampling(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments

        reads = ac.loadAlignments(testFile)

        # convert to coverage
        coverage = reads.toCoverage()

        cd1 = CoverageDistribution(self.ss, coverage, sample = 0.9)
        sum1 = sum(map(lambda x: x[1], cd1.collectedCounts.popitem()[1]))
        cd2 = CoverageDistribution(self.ss, coverage, sample = 1.0)
        sum2 = sum(map(lambda x: x[1], cd2.collectedCounts.popitem()[1]))

        # estimated counts should be around real counts
        dev = 100
        assert(sum1 > sum2 - dev and sum1 < sum2 + dev)



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

r"""
=================
CountDistribution
=================
.. currentmodule:: bdgenomics.mango.distribution
.. autosummary::
   :toctree: _generate/

   CountDistribution
   HistogramDistribution
"""

import collections
import matplotlib.pyplot as plt
import numpy as np
plt.rcdefaults()

class CountDistribution:
    """ Abstract CountDistribution class.
    Plotting functionality for visualizing count distributions of multi-sample cohorts.
    """

    #: SparkSession
    ss = None
    #: RDD of the form ((id: string, location: long), count: int))
    rdd = None
    #: fraction to sample rdd. Value should be between 0.0 and 1.0.
    sample = 1.0
    #: Whether rdd has already been sampled.
    pre_sampled = False


    def __init__(self):
        """
        Initializes a Distribution class.
        Computes the distribution of an rdd with records of the form (key: (sample ID, count), value: numObservations).
        Length is usually just a 1, and is used for reduceByKey().
        """

        # sample must be between 0 and 1
        if self.sample <= 0 or self.sample > 1:
            raise Exception('sample {} should be > 0 and <= 1'.format(self.sample))

        # sample RDD if sample is specified AND rdd has not been pre-sampled
        if self.sample < 1 and not self.pre_sampled:
            self.rdd = self.rdd.sample(False, self.sample)

        # Assign each RDD with counter. Reduce and collect.
        collectedCounts = self.rdd.reduceByKey(lambda x,y: x+y) \
            .collect()  # (id, count), number of times that count appears)

        # function that re-calculates coverage based on sampling
        approximateCounts = lambda counts, sample: int(counts * 1.0/sample)

        # restructure each record so record structure is  (key: sampleId, value: (coverage, count))
        x = list(map(lambda x: (x[0][0], (x[0][1], approximateCounts(x[1], self.sample))), collectedCounts))

        # create dictionary where keys are the sampleId
        self.collectedCounts = collections.defaultdict(set)
        for k, v in x:
            self.collectedCounts[k].add(v)


    def plotDistributions(self, normalize = True, cumulative = False, testMode = False, **kwargs):
        """
        Plots final distribution values and returns the plotted distribution as a Counter object.
        Args:
            :param normalize: normalizes readcounts to sum to 1
            :param cumulative: plots CDF of reads
            :param testMode: if true, does not generate plot. Used for testing.
            :param **kwargs: can hold figsize
        Returns:
            matplotlib axis to plot and computed data
        """

        countDistributions = {}

        if (not testMode): # For testing: do not run plots if testMode
            figsize = kwargs.get('figsize',(10, 5))
            bar_plt = kwargs.get('bar', False)
            f, ax = plt.subplots(figsize=figsize)

        # iterate through each sample
        for label, data in self.collectedCounts.items():

            if data == set(): # Do not plot any empty sets of data
                continue

            sData = sorted(data)
            values = list(map(lambda p: p[0], sData))
            counts   = list(map(lambda p: p[1], sData))

            if normalize:
                # replace distribution counts with normalized values
                sumCounts = float(sum(counts))
                counts = [i/sumCounts for i in counts]

            if cumulative:
                # calculate cumulative sum of counts
                counts = np.cumsum(counts)

            # re-write manipulated data
            countDistributions[label]=list(zip(values, counts))

            if (not testMode): # For testing: do not run plots if testMode
                if (bar_plt):
                    ax.bar(values, counts, 1, label = label)
                else:
                    ax.plot(values, counts, label = label)

        # return plots once all samples have been added
        if (not testMode): # For testing: do not run plots if testMode
            ax.legend(loc=2, shadow = True, bbox_to_anchor=(1.05, 1))
            return ax, countDistributions
        else:
            return None, countDistributions



class HistogramDistribution:
    """ Abstract HistogramDistribution class.
    Plotting functionality for visualizing count distributions of multi-sample cohorts.
    HistogramDistribution is based off of distributions with a single key.
    """

    #: SparkSession
    ss = None
    #: RDD of the form (id: string, count: int)
    rdd = None
    #: fraction to sample rdd. Value should be between 0.0 and 1.0.
    sample = 1.0
    #: Whether rdd has already been sampled.
    pre_sampled = False


    def __init__(self):
        """
        Initializes a Distribution class.
        Computes the distribution of an rdd with records of the form (key: (sample ID, count), value: numObservations).
        Length is usually just a 1, and is used for reduceByKey().
        """

        # sample must be between 0 and 1
        if self.sample <= 0 or self.sample > 1:
            raise Exception('sample {} should be > 0 and <= 1'.format(self.sample))

        # sample RDD if sample is specified AND rdd has not been pre-sampled
        if self.sample < 1 and not self.pre_sampled:
            self.rdd = self.rdd.sample(False, self.sample)

        # Assign each RDD with counter. Reduce and collect.
        collectedCounts = self.rdd.reduceByKey(lambda x,y: x+y) \
            .collect()  # id, number of times that id appears)

        # function that re-calculates coverage based on sampling
        approximateCounts = lambda counts, sample: int(counts * 1.0/sample)

        # approximate counts based on sampling
        self.collectedCounts = map(lambda x: approximateCounts(x[1], self.sample), collectedCounts)

    def plotDistributions(self, normalize = True, cumulative = False, testMode = False, **kwargs):
        """
        Plots final distribution values and returns the plotted distribution as a Counter object.

        Args:
            :param normalize: normalizes readcounts to sum to 1
            :param cumulative: plots CDF of reads
            :param testMode: if true, does not generate plot. Used for testing.
            :param **kwargs: can hold figsize

        Returns:
            matplotlib axis to plot and computed data
        """

        if (not testMode): # For testing: do not run plots if testMode
            figsize = kwargs.get('figsize',(10, 5))
            bins = kwargs.get('bins',100)

            f, ax = plt.subplots(figsize=figsize)
            ax.hist(self.collectedCounts, bins)
            return ax, self.collectedCounts
        else:
            return None, self.collectedCounts



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from bdgenomics.mango.test import SparkTestCase
from bdgenomics.mango.alignments import *

from bdgenomics.adam.adamContext import ADAMContext


class AlignmentTest(SparkTestCase):

    def test_visualize_alignments(self):

        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")

        # read alignments
        reads = ac.loadAlignments(testFile)

        alignmentViz = AlignmentSummary(self.ss, ac, reads)

        contig = "16"
        start = 26472780
        end = 26482780

        x = alignmentViz.viewPileup(contig, start, end)
        assert(x != None)

    def test_indel_distribution(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments
        reads = ac.loadAlignments(testFile)

        bin_size = 10000000
        summary = AlignmentSummary(self.ss, ac, reads)

        indels = summary.getIndelDistribution(bin_size=10000000)

        _, mDistribution = indels.plot(testMode = True, plotType="M")
        expectedM =  Counter({('1', 16 * bin_size): 225, ('1', 24 * bin_size): 150, ('1', 18 * bin_size): 150, ('1', 2 * bin_size): 150, \
                              ('1', 23 * bin_size): 150, ('1', 1 * bin_size): 75, ('1', 0 * bin_size): 75, ('1', 15 * bin_size): 75, ('1', 20 * bin_size): 75, \
                              ('1', 19 * bin_size): 75, ('1', 5 * bin_size): 75, ('1', 10 * bin_size): 75, ('1', 3 * bin_size): 75, ('1', 8 * bin_size): 75})
        assert(mDistribution == expectedM)

        _, iDistribution = indels.plot(testMode = True, plotType="I")
        expectedI =  Counter({('1', 1 * bin_size): 0, ('1', 0 * bin_size): 0, ('1', 15 * bin_size): 0, ('1', 20 * bin_size): 0, \
                              ('1', 19 * bin_size): 0, ('1', 24 * bin_size): 0, ('1', 18 * bin_size): 0, ('1', 16 * bin_size): 0, ('1', 5 * bin_size): 0,
                              ('1', 10 * bin_size): 0, ('1', 3 * bin_size): 0, ('1', 8 * bin_size): 0, ('1', 2 * bin_size): 0, ('1', 23 * bin_size): 0})
        assert(iDistribution == expectedI)

    def test_indel_distribution_maximal_bin_size(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments
        reads = ac.loadAlignments(testFile)

        summary = AlignmentSummary(self.ss, ac, reads)

        indels = summary.getIndelDistribution(bin_size=1000000000)

        _, mDistribution = indels.plot(testMode = True, plotType="M")
        expectedM =  Counter({('1', 0): 1500})
        assert(mDistribution == expectedM)


    def test_indel_distribution_no_elements(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments
        reads = ac.loadAlignments(testFile)

        summary = AlignmentSummary(self.ss, ac, reads, sample=0.00001)

        indels = summary.getIndelDistribution(bin_size=1000000000)

        _, dDistribution = indels.plot(testMode = True, plotType="D")
        expectedD =  Counter()
        assert(dDistribution == expectedD)

    def test_coverage_distribution(self):
        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments
        reads = ac.loadAlignments(testFile)

        summary = AlignmentSummary(self.ss, ac, reads)

        coverage = summary.getCoverageDistribution(bin_size = 1)
        _, cd = coverage.plotDistributions(testMode = True, cumulative = False, normalize=False)

        # first sample
        items = list(cd.popitem()[1])
        assert(len(items) == 1)
        x = items.pop()
        assert(x[0] == 1) # all positions have coverage of 1
        assert(x[0] == 1) # all positions have coverage of 1
        assert(x[1] == 1500)


    def test_fragment_distribution(self):

        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments
        reads = ac.loadAlignments(testFile)

        summary = AlignmentSummary(self.ss, ac, reads)

        fragments = summary.getFragmentDistribution()
        _, cd = fragments.plotDistributions(testMode = True, cumulative = False, normalize=False)

        # first sample
        items = list(cd.popitem()[1])
        assert(len(items) == 1)
        x = items[0]
        assert(x[0] == 75)
        assert(x[1] == 20) # all 20 sequences have same length of 75

    def test_mapq_distribution(self):

        # load file
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("small.sam")
        # read alignments
        reads = ac.loadAlignments(testFile)

        summary = AlignmentSummary(self.ss, ac, reads)

        mapq = summary.getMapQDistribution()
        _, md = mapq.plotDistributions(testMode = True, cumulative = False, normalize = False)

        # first sample
        items = list(md.popitem()[1])
        assert(len(items) == 5) # mapq for 24, 28, 35, 40, 60
        x = items[0]
        assert(x[0] == 24) # 1 read with mapq 24
        assert(x[1] == 1)



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from bdgenomics.mango.test import SparkTestCase

from bdgenomics.adam.adamContext import ADAMContext


class NotebookTest(SparkTestCase):

    def test_example(self):
        # these variables are read into mango-python.py
        spark = self.ss
        testMode = True
        alignmentFile = self.exampleFile("chr17.7500000-7515000.sam.adam")
        variantFile = self.exampleFile("snv.chr17.7502100-7502500.vcf")
        featureFile = self.exampleFile("chr17.582500-594500.bed")

        # this file is converted from ipynb in make test
        testFile = self.exampleFile("notebooks/mango-viz.py")
        exec(open(testFile).read())

    def test_coverage_example(self):
        # these variables are read into mango-python.py
        spark = self.ss
        testMode = True
        alignmentFile = self.exampleFile("chr17.7500000-7515000.sam.adam")

        # this file is converted from mango-python.coverage.ipynb in the Makefile
        testCoverageFile = self.exampleFile("notebooks/mango-python-coverage.py")
        exec(open(testCoverageFile).read())

    def test_alignment_example(self):
        # these variables are read into mango-python.py
        spark = self.ss
        testMode = True
        alignmentFile = self.exampleFile("chr17.7500000-7515000.sam.adam")

        # this file is converted from mango-python-alignment.ipynb in the Makefile
        testAlignmentFile = self.exampleFile("notebooks/mango-python-alignment.py")
        exec(open(testAlignmentFile).read())

    def test_variants_example(self):
        # these variables are read into mango-python.py
        spark = self.ss
        testMode = True
        vcfFile = self.exampleFile("genodata.v3.vcf")

        # this file is converted from mango-python-alignment.ipynb in the Makefile
        testVariantFile = self.exampleFile("notebooks/mango-python-variants.py")
        exec(open(testVariantFile).read())



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import find_packages, setup
from version import version as mango_version

try: # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements

import os

# Utility function to read the README file.
# Used for the long_description.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements('requirements.txt', session='hack')

reqs = [str(ir.req) for ir in install_reqs]

setup(
    name='bdgenomics.mango',
    version=mango_version,
    description='A scalable genomic visualization tool',
    author='Alyssa Morrow',
    author_email='akmorrow@berkeley.edu',
    url="https://github.com/bdgenomics/mango",
    install_requires=reqs,
    dependency_links=[
        'https://test.pypi.org/simple/bdgenomics-adam/'
    ],
    long_description=read('README.md'),
    packages=find_packages(exclude=['*.test.*']))



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

r"""
========
Variants
========
.. currentmodule:: bdgenomics.mango.variants
.. autosummary::
   :toctree: _generate/

   VariantSummary
"""

import bdgenomics.mango.pileup as pileup
from bdgenomics.mango.pileup.track import *
from bdgenomics.adam.adamContext import ADAMContext
from .utils import *

class VariantSummary(object):
    """ VariantSummary class.
    VariantSummary provides scrollable visualization of variants based on genomic regions.
    """

    def __init__(self, ac, rdd):
        """
        Initializes a GenomicRDD viz class.
        """
        self.ac = ac
        self.rdd = rdd

    # Takes a bdgenomics.adam.VariantContextRDD and visualizes results
    def viewPileup(self, contig, start, end, reference = 'hg19', label = "Variants", showPlot = True):
        """
        Visualizes a portion of this VariantRDD in a scrollable pileup widget

        Args:
            :param contig: contig of locus to view
            :param start: start position of locus to view
            :param end: end position of locus to view
            reference: genome build. Default is hg19
            label: name of variant track
            showPlot: Disables widget, used for testing. Default is true.

        Returns:
            pileup view for variants
        """
        contig_trimmed = contig.lstrip(CHR_PREFIX)

        # Filter RDD
        filtered = self.rdd.transform(lambda r: r.filter(((r.contigName == contig) | (r.contigName == contig_trimmed))
                                                           & (r.start < end) & (r.end > start)))

        # convert to GA4GH JSON to be consumed by mango-viz module
        json = self.ac._jvm.org.bdgenomics.mango.converters.GA4GHutil.variantRDDtoJSON(filtered._jvmRdd)

        # visualize
        if (showPlot):
            # make variant track
            tracks=[Track(viz="variants", label=label, source=pileup.sources.GA4GHVariantJson(json))]
            locus="%s:%i-%i" % (contig, start, end)
            return pileup.PileupViewer(locus=locus, reference=reference, tracks=tracks)




#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from bdgenomics.mango.test import SparkTestCase
from bdgenomics.mango.genotypes import *

from bdgenomics.adam.adamContext import ADAMContext


class GenotypesTest(SparkTestCase):

    def test_VariantsPerSampleDistribution(self):
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("genodata.v3.test.vcf")

        genotypes = ac.loadGenotypes(testFile)
        _, data = VariantsPerSampleDistribution(self.ss, genotypes).plotDistributions(testMode= True)

        expected = [6, 8, 8, 1, 7, 8]
        assert(sum(data) == sum(expected))


    def test_VariantsPerSampleDistributionSampling(self):
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("genodata.v3.test.vcf")

        genotypes = ac.loadGenotypes(testFile)
        _, data = VariantsPerSampleDistribution(self.ss, genotypes, sample=0.9).plotDistributions(testMode= True)

        expected = [6, 8, 8, 1, 7, 8]

        # estimated counts should be around real counts
        dev = 8
        assert(sum(expected) > sum(data) - dev and sum(expected) < sum(data) + dev)


    def test_GenotypeSummary(self):
        ac = ADAMContext(self.ss)
        testFile = self.resourceFile("genodata.v3.test.vcf")

        genotypes = ac.loadGenotypes(testFile)
        gs =  GenotypeSummary(self.ss, ac, genotypes)

        _, data = gs.getVariantsPerSampleDistribution().plotDistributions(testMode= True)

        expected = [6, 8, 8, 1, 7, 8]
        assert(sum(data) == sum(expected))



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

r"""
==========
Alignments
==========
.. currentmodule:: bdgenomics.mango.alignments
.. autosummary::
   :toctree: _generate/

   AlignmentSummary
   FragmentDistribution
   MapQDistribution
   IndelDistribution

"""

import bdgenomics.mango.pileup as pileup
from bdgenomics.mango.pileup.track import *
from bdgenomics.adam.adamContext import ADAMContext

from .coverage import *
from .utils import *

from collections import Counter, OrderedDict
from .distribution import CountDistribution
from cigar import Cigar
import matplotlib.pyplot as plt
plt.rcdefaults()

class AlignmentSummary(object):
    """AlignmentSummary class.
    AlignmentSummary provides scrollable visualization of alignments based on genomic regions.
    """

    def __init__(self, spark, ac, alignmentRdd, sample = 1.0):
        """
        Initializes an AlignmentSummary class.

        Args:
            :param spark: SparkSession
            :param ac: ADAMContext
            :param alignmentRdd: AlignmentRecordRDD
            :param sample: fraction of reads to sample from
        """

        # sample must be between 0 and 1
        if sample <= 0 or sample > 1:
            raise Exception('sample {} should be > 0 and <= 1'.format(self.sample))

        self.ss = spark
        self.ac = ac
        self.sample = sample
        # sample rdd
        self.rdd = alignmentRdd

        # where to store collected data
        self.coverageDistribution = None
        self.fragmentDistribution = None
        self.mapQDistribution = None
        self.indelDistribution = None

    def getCoverageDistribution(self, bin_size = 10):
        """
        Computes coverage distribution for this AlignmentRDD.

        Args:
            :param bin_size: size to bin coverage by

        Returns:
           CoverageDistribution object
        """
        if self.coverageDistribution == None:
            print("Computing coverage distributions...")
            # pre-sample before computing coverage
            sampledRdd = self.rdd.transform(lambda rdd: rdd.sample(False, self.sample))
            self.coverageDistribution = CoverageDistribution(self.ss,   \
                                                             sampledRdd.toCoverage(), \
                                                             sample = self.sample, \
                                                             bin_size = bin_size,
                                                             pre_sampled = True)

        return self.coverageDistribution

    def getFragmentDistribution(self):
        """
        Computes fragment distribution for this AlignmentRDD.

        Returns:
           FragmentDistribution object
        """
        if self.fragmentDistribution == None:
            print("Computing fragment distributions...")
            self.fragmentDistribution = FragmentDistribution(self.ss, self.rdd, sample=self.sample)

        return self.fragmentDistribution


    def getMapQDistribution(self):
        """
        Computes mapping quality distribution for this AlignmentRDD.

        Returns:
           MapQDistribution object
        """
        if self.mapQDistribution == None:
            print("Computing MapQ distributions...")
            self.mapQDistribution = MapQDistribution(self.ss, self.rdd, sample=self.sample)

        return self.mapQDistribution


    def getIndelDistribution(self, bin_size=10000000):
        """
        Computes insertion and deletion distribution for this AlignmentRDD

        Returns:
           IndelDistribution object
        """
        if self.indelDistribution == None:
            print("Computing Indel distributions...")
            self.indelDistribution = IndelDistribution(self.ss, self.rdd, bin_size=bin_size, sample=self.sample)

        return self.indelDistribution


    def viewPileup(self, contig, start, end, reference = 'hg19', label = "Reads", showPlot = True):
        """
        Visualizes a portion of this AlignmentRDD in a scrollable pileup widget

        Args:
            :param contig: contig of locus to view
            :param start: start position of locus to view
            :param end: end position of locus to view
            :param reference: genome build. Default is hg19
            :param label: name of alignment track
            :param showPlot: Disables widget, used for testing. Default is true.

        Returns:
            pileup view for alignments
        """
        contig_trimmed = contig.lstrip(CHR_PREFIX)

        # Filter RDD
        filtered = self.rdd.transform(lambda r: r.filter(((r.contigName == contig) | (r.contigName == contig_trimmed))
                                                                   & (r.start < end) & (r.end > start) & (r.readMapped)))

        # convert to GA4GH JSON to be consumed by mango-viz module
        json_map = self.ac._jvm.org.bdgenomics.mango.converters.GA4GHutil.alignmentRecordRDDtoJSON(filtered._jvmRdd)

        # visualize
        if (showPlot):
        # make pileup tracks
            tracks = []
            # iterate through all group names and make a track for each.
            for groupName in json_map:
                # create new track for this recordGroupName. files with unspecified recordGroup name
                # have a default label of "1"
                track=Track(viz="pileup", label=groupName, source=pileup.sources.GA4GHAlignmentJson(json_map[groupName]))
                tracks.append(track)

            locus="%s:%i-%i" % (contig, start, end)
            return pileup.PileupViewer(locus=locus, reference=reference, tracks=tracks)


class FragmentDistribution(CountDistribution):
    """ FragmentDistribution class.
    Plotting functionality for visualizing fragment distributions of multi-sample cohorts.
    """

    def __init__(self, ss, alignmentRDD, sample = 1.0):
        """
        Initializes a FragmentDistribution class.
        Computes the fragment distribution of a AlignmentRDD. This RDD can have data for multiple samples.

        Args:
            :param ss: global SparkSession.
            :param alignmentRDD: A bdgenomics.adam.rdd.AlignmentRDD object.
            :param sample: Fraction to sample AlignmentRDD. Should be between 0 and 1
        """

        self.sc = ss.sparkContext
        self.sample = sample
        self.rdd = alignmentRDD.toDF().rdd \
            .map(lambda r: ((r["recordGroupSample"], len(r["sequence"])), 1))

        CountDistribution.__init__(self)


class MapQDistribution(CountDistribution):
    """ MapQDistribution class.
    Plotting functionality for visualizing mapping quality distributions of multi-sample cohorts.
    """

    def __init__(self, ss, alignmentRDD, sample = 1.0):
        """
        Initializes a MapQDistribution class.
        Computes the mapping quality distribution of an AlignmentRDD. This RDD can have data for multiple samples.

        Args:
            :param ss: global SparkSession.
            :param alignmentRDD: A bdgenomics.adam.rdd.AlignmentRDD object.
            :param sample: Fraction to sample AlignmentRDD. Should be between 0 and 1
        """

        self.sc = ss.sparkContext
        self.sample = sample

        # filter out reads that are not mappped
        self.rdd = alignmentRDD.toDF().rdd \
            .filter(lambda r: (r.readMapped)) \
            .map(lambda r: ((r["recordGroupSample"], int(r["mapq"] or 0)), 1))

        CountDistribution.__init__(self)


class IndelDistribution(object):
    """ IndelDistribution class.
    IndelDistribution calculates indel distributions on an AlignmentRDD.
    """

    def __init__(self, ss, alignmentRDD, sample=1.0, bin_size=10000000):
        """
        Initializes a IndelDistribution class.
        Computes the insertiona and deletion distribution of alignmentRDD.

        Args:
            :param SparkSession: the global SparkSession
            :param alignmentRDD: A bdgenomics.adam.rdd.AlignmentRDD object
            :param bin_size: Division size per bin
        """
        bin_size = int(bin_size)
        self.bin_size = bin_size
        self.sc = ss.sparkContext
        self.sample = sample

        # filter alignments without a position
        filteredAlignments = alignmentRDD.transform(lambda x: x.sample(False, self.sample)) \
            .toDF().rdd.filter(lambda r: r["start"] != None)

        # Assign alignments with counter for contigs. Reduce and collect.
        mappedDistributions = filteredAlignments \
            .map(lambda r: ((r["contigName"], r["start"] - r["start"]%bin_size), \
                            Counter(dict([(y,x) for x,y in Cigar(r["cigar"]).items()])))) \
            .reduceByKey(lambda x,y: x+y)

        self.alignments = mappedDistributions.collect()

    def plot(self, testMode = False, plotType="I", **kwargs):
        """
        Plots final distribution values and returns the plotted distribution as a counter object.

        Args:
            :param xScaleLog: rescales xaxis to log
            :param yScaleLog: rescales yaxis to log
            :param testMode: if true, does not generate plot. Used for testing.
            :param plotType: Cigar type to plot, from ['I', 'H', 'D', 'M', 'S']

        Returns:
            matplotlib axis to plot and computed data
        """
        chromosomes = Counter()


        # count contig type at each location
        for index, counts in self.alignments:
            chromosomes[index] += counts[plotType]

        if (not testMode): # For testing: do not run plots if testMode
            # get distinct chromosome names for plot
            keys = sorted(list(set(map(lambda x: x[0][0], self.alignments))))

            # offset for max chr
            offset = 0

            # holds xtick information
            midPoints = []

            # make figure plot
            figsize = kwargs.get('figsize',(10, 5))
            f, ax = plt.subplots(figsize=figsize)

            # for all chromosomes
            for key in keys:

                # filter by chromosome key
                values = [x for x in chromosomes.items() if x[0][0] == key]

                # get positions
                positions = map(lambda x: x[0][1] + offset, values)

                # get counts
                counts = map(lambda x: x[1], values)

                ax.plot(positions, counts, ls='', marker='.')

                # set label for chromosome
                midPoint = min(positions) + (max(positions) - min(positions))/2
                midPoints.append(midPoint)

                offset = max(positions)


            ax.set_xticks(midPoints)
            ax.set_xticklabels(keys, rotation=-90, size=8.5)
            return ax, chromosomes

        else:
            return None, chromosomes



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

r"""
========
Coverage
========
.. currentmodule:: bdgenomics.mango.coverage
.. autosummary::
   :toctree: _generate/

   CoverageDistribution
"""

import collections
import matplotlib.pyplot as plt
import numpy as np
from .distribution import CountDistribution
plt.rcdefaults()

class CoverageDistribution(CountDistribution):
    """ CoverageDistribution class.
    Plotting functionality for visualizing coverage distributions of multi-sample cohorts.
    """

    def __init__(self, ss, coverageRDD, sample = 1.0, name="coverage", bin_size = 10, pre_sampled = False):
        """
        Initializes a CoverageDistribution class.
        Computes the coverage distribution of a CoverageRDD. This RDD can have data for multiple samples.

        Args:
            :param ss: global SparkSession.
            :param coverageRDD: A bdgenomics.adam.rdd.CoverageRDD object.
            :param sample: Fraction to sample CoverageRDD. Should be between 0 and 1
            :param name: sample name

        """

        self.sc = ss.sparkContext
        self.sample = sample
        # TODO replace name with r["sampleId"], dependent on https://github.com/bigdatagenomics/adam/pull/2051
        self.rdd = coverageRDD.toDF().rdd \
            .map(lambda r: ((name, r["count"] - r["count"]%bin_size), (int(r["end"])-int(r["start"]))))

        CountDistribution.__init__(self)



#
# Licensed to Big Data Genomics (BDG) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The BDG licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import os
import sys
import tempfile
import unittest


from pyspark.sql import SparkSession

class SparkTestCase(unittest.TestCase):


    def resourceFile(self, file):

        mangoRoot = os.path.dirname(os.getcwd())
        return os.path.join(os.path.join(mangoRoot, "mango-core/src/test/resources"), file)


    def exampleFile(self, file):

        mangoRoot = os.path.dirname(os.getcwd())
        return os.path.join(os.path.join(mangoRoot, "example-files"), file)


    def tmpFile(self):

        tempFile = tempfile.NamedTemporaryFile(delete=True)
        tempFile.close()
        return tempFile.name


    def checkFiles(self, file1, file2):

        f1 = open(file1)
        f2 = open(file2)

        try:
            self.assertEquals(f1.read(), f2.read())
        finally:
            f1.close()
            f2.close()


    def setUp(self):
        self._old_sys_path = list(sys.path)
        class_name = self.__class__.__name__
        self.ss = SparkSession.builder.master('local[4]').appName(class_name).getOrCreate()
        self.sc = self.ss.sparkContext


    def tearDown(self):
        self.ss.stop()
        sys.path = self._old_sys_path
